#include <cmath>
#include <iostream>
#include <memory>

void print_indent(int indent) {
	for(; indent>0; indent--) {
		std::cout << "  ";
	}
}


class Operator {
  public:
	Operator() = default;
	virtual ~Operator() = default;

	virtual float compute()const = 0;
	virtual void print(int indent)const = 0;

  protected:
	Operator(const Operator&) = default;
	Operator(Operator&&) = default;
	Operator& operator=(const Operator&) = default;
	Operator& operator=(Operator&&) = default;
};

class Reference final : public Operator {
  public:
	Reference(const float& value) : value_(value) {}
	
	float compute()const override {
		return value_;
	}
	
	void print(int indent)const override {
		print_indent(indent);
		std::cout << "[" << value_ << "]\n";
	}

  private:
	const float& value_;
};


// TODO


auto value(float& v) {
	return std::make_unique<Reference>(v);
}
auto operator+(std::unique_ptr<Operator> lhs, std::unique_ptr<Operator> rhs) {
	return std::make_unique<Add>(std::move(lhs), std::move(rhs));
}
auto operator-(std::unique_ptr<Operator> lhs, std::unique_ptr<Operator> rhs) {
	return std::make_unique<Subtract>(std::move(lhs), std::move(rhs));
}
auto operator*(std::unique_ptr<Operator> lhs, std::unique_ptr<Operator> rhs) {
	return std::make_unique<Multiply>(std::move(lhs), std::move(rhs));
}
auto operator/(std::unique_ptr<Operator> lhs, std::unique_ptr<Operator> rhs) {
	return std::make_unique<Divide>(std::move(lhs), std::move(rhs));
}

int main() {
	auto a = 1.f;
	auto b = 1.f;
	auto c = -1.f;
	auto equation = (value(a) + value(b)/value(b) - value(c))
	              / (value(a) * value(a));

	std::cout << equation->compute() << "\n";

	a = 2.f;
	b = -9.f;
	c = 2.f;
	std::cout << equation->compute() << "\n\n";

	equation->print(0);
}

